(******************************************************************************
**
**  Object in this unit was originally written by and copyright
**                     by Andy Strong - Compuserve ID   : 100716,3015
**                                      Internet Address: andrews@nbaqsl.co.uk
**
**  Released into the public domain 14/02/97
**
**  Controls:
**  This code may be used, modified, included in applications without any
**  license agreements as long as the disclaimers are accepted, and the
**  comments are left intact ( Some Hope! <g> ),
**
**  Disclaimer:
**  This software is released into the public domain on the strict understanding
**  that neither myself nor any associates or companies I work for have any
**  liability explictly or implied.
**
**  STAT_DLG
**
**  Status Dialogue unit.  Will generate a Status form without the use of
**  a DFM file.  Allows a Status Panel, Status Gauge or both.
**  Originally written to provide progress status feedback whilst PACKing
**  a table.
**
*******************************************************************************)
unit Stat_Dlg;

interface

uses
  Forms, ExtCtrls, Gauges;

type
  TStatusType = (stPanel, stProgressBar, stBoth);

  TStatusDialogue = class(TForm)      { Typical Status Dialogue }
    StatusPanel: TPanel;
    StatusGauge: TGauge;
  public
    { Our Own Special constructor type procedure - negates the need for DFM file }
    class procedure CreateStatusDialogue(var FSForm: TStatusDialogue;
                    StatType: TStatusType; WinCaption: openstring);
    destructor Destroy; override;     { Used to free components memory }
  end;

implementation

uses
  Graphics, Controls;

(*
**  Routine to create a PACK dialogue status form 'on-the-fly'
*)
class procedure TStatusDialogue.CreateStatusDialogue(var FSForm: TStatusDialogue;
                StatType: TStatusType; WinCaption: openstring);
begin
  { Dynamically Create and show the Extra Dialogue - experiment!}
  FSForm := TStatusDialogue.CreateNew(Application);
  try
    with FSForm do
    begin
      ClientWidth := 425;
      ClientHeight := 50;
      Top := 132;
      Left := 233;
      FormStyle := fsStayOnTop;
      BorderIcons := [];
      BorderStyle := bsDialog;
      Position := poDesigned;
      Caption := WinCaption;
      ControlStyle := ControlStyle + [csOpaque];

      StatusPanel := nil;
      StatusGauge := nil;

      if (StatType = stPanel) or (StatType = stBoth) then
      begin
        { Add Status Panel component }
        StatusPanel := TPanel.Create(FSForm);
        with StatusPanel do
        begin
          Parent := FSForm;
          Height := 22;
          Width := 408;
          BevelInner := bvLowered;
          Top := 12;
          Left := 8;
          ControlStyle := ControlStyle + [csOpaque];
        end;
      end;
      if (StatType = stProgressBar) or (StatType = stBoth) then
      begin
        { Add Status Gauge component }
        StatusGauge := TGauge.Create(FSForm);
        with StatusGauge do
        begin
          Parent := FSForm;
          Height := 22;
          Width := 408;
          Left := 8;
          ControlStyle := ControlStyle + [csOpaque];
          if StatType = stBoth then
          begin
            FSForm.ClientHeight :=
              FSForm.ClientHeight +
              StatusPanel.Height +
              (StatusPanel.Height div 2);
            Top := 12 + StatusPanel.Height +
              (StatusPanel.Height div 2);
          end else
            Top := 12;
          Progress := 0;
          ForeColor := clBlue;
        end;
      end;
    end;
  except
    FSForm.Free;                      { Frees ALL created components on error }
    FSForm := nil;
  end;
end;

(*
**  Destructor for PackStatus Dialogue.  Releases memory taken by statuspanel
*)
destructor TStatusDialogue.Destroy;
begin
  StatusPanel.Free;
  StatusGauge.Free;
  inherited Destroy;
end;

end.
