(******************************************************************************
**
**  Object in this unit was originally written by and copyright
**                     by Andy Strong - Compuserve ID   : 100716,3015
**                                      Internet Address: andrews@nbaqsl.co.uk
**
**  Released into the public domain 14/02/97
**
**  Controls:
**  This code may be used, modified, included in applications without any
**  license agreements as long as the disclaimers are accepted, and the
**  comments are left intact ( Some Hope! <g> ),
**
**  Disclaimer:
**  This software is released into the public domain on the strict understanding
**  that neither myself nor any associates or companies I work for have any
**  liability explictly or implied.
**
**  TDIB_SRC.PAS
**
**  Unit to encapsulate the handling of Device Independent Bitmaps (DIBS)
**  I wrote this unit to get around problems with Delphi 1.0 tying the TBitmap
**  object to the screen colour depth.  This is fine as long the system is not
**  upgraded.
**
**  NOTE: JPEG handling removed due to copyright reasons.
**
*******************************************************************************)
unit Tdib_src;

interface
{.$DEFINE MEMMOND}
{$DEFINE JPEGHandler}
uses
  Classes, WinTypes, SysUtils, Graphics, Gauges, Forms, DB, Stat_Dlg, ExtCtrls;

type
  { Currently supported Image types }
  TImageType = (itBMP{$ifdef JPEGHandler}, itJPEG{$endif});

  PRGBQList = ^TRGBQList;
  TRGBQList = Record
    RGBQIndex: array[0..0] of TRGBQuad;
  end;
  {$ifdef MEMMOND}TMark = procedure(const Msg : String);{$endif}

  EDIBException = class(Exception);   { DIB Exceptions }

  TDIB = class                        { DIB handler }
  private
    { Properties }
    FDIBPointer: Pointer;             { Pointer to Memory DIB }
    FDIBHandle: THandle;              { Global Memory Handle }
    FDIBSize: LongInt;                { Physical Size of memory for DIB }
    FDIBHeight: integer;              { DIB Height - Scan lines }
    FDIBWidth: integer;               { DIB Width }
    FDIBColourBits: byte;             { Number of Colour Bits - 1, 4, 8 or 24 }
    FSysColBits: byte;                { Number of colour bits system supports }
    FMemHandle: THandle;              { Temporary Global Mem Handle for 256 DIB }
    FStatusForm: TStatusDialogue;     { Progress Status Dialogue }
    FJPEGQuality: byte;               { Quality image percentage retention }

    { Methods }
    function Align_DWord(Address: LongInt): LongInt;
    function CopyToClipboard(DIBPtr: Pointer): boolean;
    {$ifdef JPEGHandler}
    function CreateDIBFromJPEGStream(Stream: TStream): boolean;
    {$endif}
    function CreateDIBFromStream(Stream: TStream): boolean;
    function ExpandTo24Bits(SrcPtr: Pointer; Progress: boolean): THandle;
    function Get256ColDIBFrame(const BWidth, BHeight: integer): Pointer;
    function GetBMPBitsPointer: Pointer;
    function GetBMPInfoPointer: PBitmapInfoHeader;
    function GetDIBEmpty: boolean;
    function GetPaletteLength(DIBPtr: TBitMapInfoHeader): Integer;
    function GetSystemColours: word;
    function IncreaseTo24bit(DIBPointer: Pointer; StatusBar: boolean): THandle;
    function MakeQDefaultPaletteArray: Pointer;
    function ReduceTo256Cols(StatusBar: boolean): Pointer;
    function GetDIBPointer: Pointer;

    procedure ClearInternals;
    procedure ConvertRGBLine(RGBBuffer, RGBBuffer2, DestPointer: Pointer;
                             LineWidth: integer);
    procedure CreateBMPFromDIB(DIBPtr: Pointer; var TBMP: TBitMap);
    procedure Release256Frame;
    procedure ReleaseDIB;
    procedure SetDIBHandle(GlobalHandle: THandle);
    procedure SetInternals(DIBPtr: Pointer; SetSize: boolean);
    {$ifdef JPEGHandler}
    procedure SetJPEGQuality(Value: byte);
    procedure WriteJPEGToStream(Stream: TStream; DIBPointer: Pointer);
    {$endif}
    procedure WriteDIBToStream(Stream: TStream; DIBPointer: Pointer);

  public
    { Methods }
    constructor Create;
    constructor CreateFromBlobField(BlobField: TField);
    constructor CreateFromClipboard;
    constructor CreateFromMemory(GlobalHandle: THandle);
    constructor CreateFromStream(Stream: TStream);
    destructor Destroy; override;

    class procedure CreateMetaImageFromFile(FName: TFileName; var TargetImage: TImage);
    class procedure ConvertDIBToMetaFile(DIB: TDIB; TargetBox: TPoint;
                                          var MFile: TMetaFile);
    function CopyDIBToClipboard: boolean;
    function GetDIBFromClipBoard: boolean;
    function MakeDIBFromStream(Stream: TStream): boolean;

    procedure Clear;
    procedure ConvertTo24Bit(StatusBar: boolean);
    procedure ConvertTo8Bit(StatusBar: boolean);
    procedure ReleaseDIBLock;
    procedure MakeDelphiBMPFromDIB(TBmp: TBitmap; ForceStatusBar: boolean);
    procedure MakeDIBFromBlobStream(BlobField: TField);
    procedure PutDIBIntoStream(Stream: TStream; IType: TImageType);

    { Properties }
    property DIBBitsPointer: Pointer read GetBMPBitsPointer;
    property DIBColourBits: byte read FDIBColourBits;
    property DIBHeaderPointer: Pointer read GetDIBPointer;
    property DIBInfoPointer: PBitmapInfoHeader read GetBMPInfoPointer;
    property DIBHeight: integer read FDIBHeight;
    property DIBWidth: integer read FDIBWidth;
    property DIBEmpty: boolean read GetDIBEmpty;
    {$ifdef JPEGHandler}
    property JPEGQuality: byte read FJPEGQuality write SetJPEGQuality;
    {$endif}
    property SystemColourBits: byte read FSysColBits;
  end;

implementation

uses
  Dialogs, WinProcs, ClipBrd, AjsTools, Controls, DBTables, ToolHelp
  {$ifdef JPEGHandler}, DlphJPEG{$endif};

{$I+}

{$ifdef MEMMOND}
  procedure Mark(const S : String);
  var
    M : TMark;
  begin
    @M := GetProcAddress(GetModuleHandle('MEMMOND'),'MARK');
    if Assigned(M) then
      M(S);
  end;
{$endif}
(*
**  Define the Selector shift to be added to enable skipping of segment
**  boundaries sucessfully
*)
procedure __AHSHIFT; far; external 'KERNEL' index 113;

(*
**  Routine to enable a Memory pointer to be incremented by 'Ofs' value
**  whilst handling the skipping of Segment boundaries
*)
function OffsetPointer(P: Pointer; Ofs: Longint): Pointer; assembler;
asm
  MOV     AX,Ofs.Word[0]
  MOV     DX,Ofs.Word[2]
  ADD     AX,P.Word[0]
  ADC     DX,0
  MOV     CX,OFFSET __AHSHIFT
  SHL     DX,CL
  ADD     DX,P.Word[2]
end;

(*
**  Constructor.  Initialises variables, creates objects
*)
constructor TDIB.Create;
begin
  inherited Create;
  {$ifdef MEMMOND}  Mark('TDIB Constructor'); {$endif}
  ClearInternals;
  FSysColBits := GetSystemColours;
end;

(*
**  DIBs are held in TBitmapFileInformation format on Blob fields. This procedure
**  opens a stream to the DIB in the database, and then calls the procedure to
**  create the DIB from the stream.
*)
constructor TDIB.CreateFromBlobField(BlobField: TField);
begin
  inherited Create;
  ClearInternals;
  FSysColBits := GetSystemColours;
  MakeDIBFromBlobStream(BlobField);
end;

(*
**  Constructor.  Initialises variables, creates DIB from passed Stream
*)
constructor TDIB.CreateFromStream(Stream: TStream);
begin
  inherited Create;
  {$ifdef MEMMOND}  Mark('TDIB Constructor'); {$endif}
  ClearInternals;
  FSysColBits := GetSystemColours;
  if not MakeDIBFromStream(Stream) then
    raise EDIBException.Create('Invalid DIB format in stream');
end;

(*
**  Constructor.  Initialises variables, creates DIB from Clipboard
*)
constructor TDIB.CreateFromClipBoard;
begin
  inherited Create;
  {$ifdef MEMMOND}  Mark('TDIB Constructor'); {$endif}
  ClearInternals;
  FSysColBits := GetSystemColours;
  if Clipboard.HasFormat(CF_DIB) then
    GetDIBFromClipBoard
  else
    raise EDIBException.Create('No DIB on the Clipboard');
end;

(*
**  Constructor.  Initialises variables, 'creates' DIB from global memory
**  handle containing a DIB already.
*)
constructor TDIB.CreateFromMemory(GlobalHandle: THandle);
var
  DIBPtr: PBitmapInfoHeader;
begin
  inherited Create;
  {$ifdef MEMMOND}  Mark('TDIB Constructor'); {$endif}
  ClearInternals;
  FSysColBits := GetSystemColours;
  SetDIBHandle(GlobalHandle);               { Make this our DIB }
end;


(*
**  Destructor.  Frees up grabbed system resources
*)
destructor TDIB.Destroy;
begin
  ReleaseDIB;
  {$ifdef MEMMOND}  Mark('TDIB Destructor'); {$endif}
  inherited Destroy;
end;

(*
**  DIBs are held in either TBitmapFileInformation format or JPEG on Blob
**  fields. This procedure opens a stream to the DIB in the database,
**  and then calls the procedure to create the DIB from the stream.
*)
procedure TDIB.MakeDIBFromBlobStream(BlobField: TField);
var
  BmpFileHeader: PBitmapFileHeader;
  BlobStream: TBlobStream;
  BlobOffs: byte;
begin
  GetMem(BmpFileHeader,
    SizeOf(TBitmapFileHeader));             { Get BMP FIle Header memory }
  FillChar(BmpFileHeader^,
    SizeOf(TBitmapFileHeader), 0);          { FIll with #0 }
  BlobStream := TBlobStream.Create(
    TGraphicField(BlobField), bmRead);      { Open up Blob stream to field }
  try
    with BlobStream do
    begin
      BlobOffs := 0;                        { Should start with valid ID }
      Seek(BlobOffs, 0);
      Read(BmpFileHeader^, SizeOf(TBitmapFileHeader));
      if (BmpFileHeader^.bfType <> BMPIdentifier)
      {$ifdef JPEGHandler}
        and (BmpFileHeader^.bfType <> JPEGIdentifier)
      {$endif} then { if Valid DIB }
      begin
        BlobOffs := 8;                      { Bodge for older Databases!!!!! }
        Seek(BlobOffs, 0);
        Read(BmpFileHeader^, SizeOf(BmpFileHeader));
      end;

      if (BmpFileHeader^.bfType = BMPIdentifier)
      {$ifdef JPEGHandler}
        or (BmpFileHeader^.bfType = JPEGIdentifier)
      {$endif} then { if Valid DIB }
      begin
        Seek(BlobOffs, 0);                  { Reset to File Header }
        if not MakeDIBFromStream(BlobStream) then { Make DIB }
          raise EDIBException.Create('Could not create DIB from Blob Field');
      end;                                  { Not recognised - ignore }
    end;
  finally
    ReleaseDIBLock;
    FreeMem(BmpFileHeader, SizeOf(TBitmapFileHeader));
    BlobStream.Free;
  end;
end;

(*
**  Pastes a DIB from the clipboard straight into a Memory DIB.
**  This is to ensure that Delphi never sees it.  Delphi will corrupt a 24bit
**  DIB if running on a 256 colour system (it will make a bad attempt at
**  converting it to 8bit colour).
*)
function TDIB.GetDIBFromClipBoard: boolean;
var
  CHandl: THandle;                          { Handle to Clipboard }
  PDIB: Pointer;                            { Pointer to DIB structure }
  PGlobData: PGlobalEntry;                  { Global Memory Info Structure }
begin
  if ClipBoard.HasFormat(CF_DIB) then
  begin
    Clear;                                  { Release current Memory if any }
    GetMem(PGlobData, SizeOf(TGlobalEntry));{ Storage for Global Handle Info }
    CHandl := ClipBoard.GetAsHandle(CF_DIB);{ Get Handle to DIB }
    PDIB := GlobalLock(CHandl);             { Convert Handle to a pointer }
    try
      FillChar(PGlobData^, SizeOf(TGlobalEntry), 0);
      PGlobData^.dwSize :=
        SizeOf(TGlobalEntry);               { Set Size for pernickety toolhelp }
      GlobalEntryHandle(PGlobData, CHandl); { Get Memory size amongst others }

      FDIBHandle := GlobalAlloc(GMEM_MOVEABLE,
        PGlobData^.dwBlockSize);            { Get DIB Memory }

      hMemCpy(DIBInfoPointer, PDIB,
        PGlobData^.dwBlockSize);            { Copy Data }

      SetInternals(DIBInfoPointer, False);  { Set Pointers etc }

      FDIBSize := PGlobData^.dwBlockSize;   { DIB Size - override }

    finally
      GlobalUnlock(CHandl);                 { Release ClipBoard Handle }
      FreeMem(PGlobData, SizeOf(TGlobalEntry));
    end;
    ReleaseDIBLock;
  end;
end;

(*
**  Routine to Set Internal DIB releated variables
*)
procedure TDIB.SetInternals(DIBPtr: Pointer; SetSize: boolean);
var
  PBMPInfo: PBitmapInfo;
begin
  PBMPInfo := DIBPtr;                       { Pointer to Info block }

  with PBMPInfo^.bmiHeader do
  begin

    FDIBWidth := biWidth;                   { Save this Images size }
    FDIBHeight := biHeight;
    FDIBColourBits := biBitCount;           { Bits Per Pixel }
    if SetSize then                         { May have already been set }
      FDIBSize :=  FDIBHeight * Align_DWord({ DIB Height multiplied by }
      LongInt(FDIBWidth * biBitCount) div 8){ Bits per pixel multiplied by width }
      + SizeOf(TBitmapInfoHeader) +         { plus Header Length }
      GetPaletteLength(PBMPInfo^.bmiHeader);{ Plus palette length & Header Len }
  end;
end;

(*
**  Routine to determine the colour capability of the host system
**  Returns 1, 4, 8 or 24
*)
function TDIB.GetSystemColours: word;
var
  ScreenDC: HDC;
begin
  ScreenDC := GetDC(0);                     { Get Screen DC }
  try
    { Get the device capabilities of the screen - Colour Bits}
    Result :=
      GetDeviceCaps(ScreenDC, BITSPIXEL) *
      GetDeviceCaps(ScreenDC, PLANES);      { Get Screen colour depth bits }
  finally
    ReleaseDC(0, ScreenDC);                 { Release DC memory and handle }
  end;
  if Result < 8 then
    SoundDlg('This system is NOT enabled for a' + #13 +
      'minimum of 256 colours', mtWarning, [mbOk], 0);
end;

(*
**  Routine to Clear Internal DIB releated variables
*)
procedure TDIB.ClearInternals;
begin
  FDIBSize := 0;
  FDIBPointer := nil;
  FDIBHandle := 0;
  FDIBHeight := 0;
  FDIBWidth := 0;
  FDIBColourBits := 0;
  FMemHandle := 0;
  FStatusForm := nil;
  FJPEGQuality := 85;
end;

(*
**  Procedure to release the temporary memory used to generate a 256 colour
**  version of a 24 bit colour image
*)
procedure TDIB.Release256Frame;
begin
  if FMemHandle <> 0 then
  begin
    GlobalUnlock(FMemHandle);
    if GlobalFree(FMemHandle) <> 0 then
      raise EDIBException.Create('Unable to free 256 colour global memory');
    FMemHandle := 0;
  end;
end;

(*
**  DIBs are held by pointers to contigous blocks of memory.  This procedure
**  releases a DIBs system resources.
*)
procedure TDIB.ReleaseDIB;
begin
  if FDIBHandle <> 0 then                   { Release memory from old one }
  begin
    GlobalUnlock(FDIBHandle);               { Release Global Memory Handle }
    if GlobalFree(FDIBHandle) <> 0 then
      raise EDIBException.Create('Unable to free global memory');
    Release256Frame;                        { Release 256 colour frame memory }
    ClearInternals;
  end;
end;

(*
**  Sets Pointer to Global Memory object handle, and locks it if necessary
*)
function TDIB.GetDIBPointer: Pointer;
begin
  if not Assigned(FDIBPointer) and (FDIBHandle <> 0) then
    FDIBPointer := GlobalLock(FDIBHandle);
  Result := FDIBPointer;
end;

(*
**  Function to return a pointer to the DIBs TBitmapInfo structure in memory
*)
function TDIB.GetBMPInfoPointer: PBitmapInfoHeader;
begin
  if FDIBHandle <> 0 then
    Result := DIBHeaderPointer
  else
    Result := nil;
end;

(*
**  Function to return a pointer to the DIBs Bits structure in memory
*)
function TDIB.GetBMPBitsPointer: Pointer;
var
  TmpBMPInfo: PBitmapInfoHeader;
begin
  if FDIBHandle <> 0 then
  begin
    TmpBMPInfo := DIBHeaderPointer;
    Result := OffsetPointer(DIBHeaderPointer,
              SizeOf(TBitmapInfoHeader) + GetPaletteLength(TmpBMPInfo^));
  end
  else
    Result := nil;
end;

(*
**  Function to check whether we have a DIB loaded or not
*)
function TDIB.GetDIBEmpty: boolean;
begin
  Result := (FDIBHandle = 0);               { As Simple as that! }
end;

(*
**  Procedure to release the memory lock for windows to move memory if needed
*)
procedure TDIB.ReleaseDIBLock;
begin
  if Assigned(FDIBPointer) then
    GlobalUnlock(FDIBHandle);
  FDIBPointer := nil;
end;

(*
**  Function to clear down the current DIB from memory
*)
procedure TDIB.Clear;
begin
  ReleaseDIB;
end;

(*
**  Function to return the Palette size for a DIB pointed to by DIBPtr
*)
function TDIB.GetPaletteLength(DIBPtr: TBitMapInfoHeader): Integer;
var
  RGBSize: integer;
begin
  RGBSize := DIBPtr.biClrUsed * SizeOf(TRGBQuad);{ Size of RGB table }

  if RGBSize = 0 then
    RGBSize := (1 shl DIBPtr.biBitCount) * SizeOf(TRGBQuad);

  GetPaletteLength := RGBSize;
end;

(*
**  Creates a DIB in memory from the passed stream.  Assumes that the Stream
**  pointer is pointing at the start of a TBitmapFileInfo structure, and that
**  what follows immediately is a standard uncompressed Windows 3.0 DIB
**  Note:  Our Memory DIB does NOT contain the TBitmapFileHeader!!!!
**  Routine to decide what format the stream is in
*)
function TDIB.MakeDIBFromStream(Stream: TStream): boolean;
var
  CurrentPosn: Longint;
  IDword: word;
begin
  with Stream do
  begin
    Result := False;                        { Assume failure }
    CurrentPosn := Position;                { Save Stream position }
    Read(IDWord, SizeOf(IDWord));           { Read Identifier }
    Seek(CurrentPosn, 0);                   { Back to where we started }

    case IDWord of
      BMPIdentifier:                        { BMP in stream? }
        Result :=
          CreateDIBFromStream(Stream);      { Make DIB from BMP }
      {$ifdef JPEGHandler}
      JPEGIdentifier:                       { JPEG in stream? }
        Result :=
          CreateDIBFromJPEGStream(Stream);  { Make DIB from JPEG }
      {$endif}
      else
        raise EDIBException.Create('Invalid image format in stream');
    end;
  end;
end;

(*
**  Function to create a DIB from a DIB in the Stream
*)
function TDIB.CreateDIBFromStream(Stream: TStream): boolean;
var
  PBmpFileHeader: PBitmapFileHeader;
  PBmpInfoHeader: PBitmapInfoHeader;
  PBmpInfo: PBitmapInfo;
  CurrentPosn: Longint;
  DIBSize: LongInt;
begin
  Result := False;
  GetMem(PBmpFileHeader,
    SizeOf(TBitmapFileHeader));             { File Header memory }
  GetMem(PBmpInfoHeader,
    SizeOf(TBitmapInfoHeader));             { Info Header memory }
  try
    try
      with Stream do
      begin
        CurrentPosn := Position;            { Save Stream position }
        Read(PBmpFileHeader^, SizeOf(TBitmapFileHeader));{ Read in File Header }
        if PBmpFileHeader^.bfType = BMPIdentifier then  { Check for valid DIB }
        begin
          FDIBSize := PBmpFileHeader^.bfSize -
                    SizeOf(TBitmapFileHeader);          { DIB Size }

          Read(PBmpInfoHeader^, SizeOf(TBitMapInfoHeader));{ Read in DIB info header }

          if PBmpInfoHeader^.biCompression <> BI_RGB then
            raise EDIBException.Create('Cannot handle compressed DIBS');

          Seek(CurrentPosn + SizeOf(TBitmapFileHeader), 0);{ Point at DIB header }

          FDIBHandle := GlobalAlloc(GMEM_MOVEABLE,
            FDIBSize);                      { Allocate memory }

          if Read(DIBHeaderPointer^, FDIBSize) = FDIBSize then{ Read the DIB into Memory }
          begin
            SetInternals(DIBInfoPointer, False);
            PBmpInfo := DIBHeaderPointer;
            { Since some drivers and applications do not set the biSizeImage correctly
              or at all, we will ensure it is correct, 'cos we use it later on }
            PBmpInfo^.bmiHeader.biSizeImage :=
                      (PBmpFileHeader^.bfSize - PBmpFileHeader^.bfOffBits);
            Result := True;                 { All Okay }
          end;
        end;
      end;
    except
      ReleaseDIB;                           { Release Memory if necessary }
      raise EDIBException.Create('ERROR in creating DIB from Stream');
    end;
  finally
    FreeMem(PBmpFileHeader, SizeOf(TBitmapFileHeader)); { File Header memory }
    FreeMem(PBmpInfoHeader, SizeOf(TBitmapInfoHeader)); { Info Header memory }
  end;
  ReleaseDIBLock;                           { Allow Windows to move it }
end;

{$ifdef JPEGHandler}
(*
**  Function to create a DIB from a JPEG in the Stream
*)
function TDIB.CreateDIBFromJPEGStream(Stream: TStream): boolean;
const
  FileNameSize = 145;
var
  StrmFile: TFileStream;                    { Stream to Disk }
  JPG: TJPEG;                               { JPEG Object }
  FName: String[FileNameSize];              { Temporary file name }
  MemHandle: THandle;                       { Memory Handle }
begin
  FName := TJPEG.GetTempOutputFile;         { Get Temporary output file name }
  StrmFile := TFileStream.Create(FName,
    fmCreate);                              { Create temporary file }
  try
    StrmFile.CopyFrom(Stream, Stream.Size); { Copy stream to file }
  finally
    StrmFile.Free;                          { Closes file }
  end;

  JPG := TJPEG.Create;                      { Create JPEG object }
  try
    with JPG do
    begin
      QualityPercent := FJPEGQuality;       { Set Our Quality retention }
      ProgressSteps := 16;                  { 16 Steps }
      ProgressBar := True;                  { Show progress }
      MemHandle := Decompress(FName);       { Decompress into memory }
      SetDIBHandle(MemHandle);              { Make it ours }
    end;
  finally
    JPG.Free;                               { Release JPEG Object }
    DeleteFile(FName);                      { Remove Temporary file }
  end;
end;
{$endif}

(*
**  Procedure to write a DIB pointed to by DIBPointer to the passed stream.
**  Prepends a TBitmapFileHeader to the DIB.
*)
procedure TDIB.WriteDIBToStream(Stream: TStream; DIBPointer: Pointer);
var
  BitSize: LongInt;                         { BITS size }
  PalSize: Integer;                         { Actual Palette Size }
  PBMP: PBitmapInfoHeader;                  { Pointer to TBITMAPINFO HDR structure }
  PTmpBMI: PBitmapFileHeader;               { Bitmap file header structure }
begin
  PBMP := DIBPointer;
  GetMem(PTmpBMI, SizeOf(TBitmapFileHeader));{ File Header memory }
  try
    PalSize := GetPaletteLength(PBMP^);     { Palette Size }
    BitSize := PBMP^.biSizeImage;           { Bitmap Size }
    try
      with PTmpBMI^ do                      { Fill in Bitmap File header }
      begin
        bfType := BMPIdentifier;            { Set Bitmap File Identifier }
        bfSize := SizeOf(TBitmapFileHeader) + SizeOf(TBitMapInfoHeader) +
                    PalSize + BitSize;
        bfReserved1 := 0;
        bfReserved2 := 0;
        bfOffbits := SizeOf(TBitmapFileHeader) + SizeOf(TBitMapInfoHeader) + PalSize;
      end;

      with Stream do
      begin
        Write(PTmpBMI^, SizeOf(TBitmapFileHeader));{ Write BITMAP File Header details }
        Write(PBMP^, SizeOf(TBitmapInfoHeader) + PalSize + BitSize);
      end;
    except
      raise EDIBException.Create('Error writing DIB to Stream');
    end;
  finally
    FreeMem(PTmpBMI, SizeOf(TBitmapFileHeader));{ File Header memory }
  end;
end;

{$ifdef JPEGHandler}
(*
**  Procedure to write a DIB pointed to by DIBPointer to the passed stream
**  in JPEG format.
*)
procedure TDIB.WriteJPEGToStream(Stream: TStream; DIBPointer: Pointer);
var
  JPG: TJPEG;                               { JPEG manilulation object }
  InFile: TFileStream;                      { File Stream to read in compressed }
  TmpFile: PString;                         { Temporary FileName }
begin
  JPG := TJPEG.Create;                      { Create JPEG Object }
  New(TmpFile);
  try
    with JPG do
    begin
      QualityPercent := FJPEGQuality;       { Set Our Quality retention }
      JPG.ProgressSteps := 16;              { 16 Steps }
      ProgressBar := True;                  { Show Status bar }
      TmpFile^ := Compress(DIBPointer, ''); { Compress into temp file name}

      InFile := TFileStream.Create(TmpFile^,
        fmOpenRead);                        { Open New JPEG file }
      try
        Stream.CopyFrom(InFile, InFile.Size);{ Copy into target stream }
      finally
        InFile.Free;                        { Close JPEG file }
      end;
      DeleteFile(TmpFile^);                 { Delete JPEG File }
    end;
  finally
    JPG.Free;                               { Free JPEG Object }
    Dispose(TmpFile);
  end;
end;
{$endif}
(*
**  Streams our DIB in memory to the passed stream in the selected format.
*)
procedure TDIB.PutDIBIntoStream(Stream: TStream; IType: TImageType);
begin
  case IType of
    itBMP:
      WriteDIBToStream(Stream,
        DIBHeaderPointer);                  { Write BMP to Stream }
    {$ifdef JPEGHandler}
    itJPEG:
      WriteJPEGToStream(Stream,
        DIBHeaderPointer);                  { Write JPEG copy to stream }
    {$endif}
  end;
  ReleaseDIBLock;                           { Allow Windows to Move DIB }
end;

(*
**  DIB Lines (Source and Destinations) must always be Double word aligned
*)
function TDIB.Align_DWord(Address: LongInt): LongInt;
begin
  Result := LongInt(((Address + 3) div 4) * 4);
end;

(*
**  Function to create a default 128 colour palette with a good even spread
**  Red, Green & Blue hues.  This is so that any form that shows a picture
**  can be realized against a standard palette.
*)
function TDIB.MakeQDefaultPaletteArray: Pointer;
var
  QRGBList: PRGBQList;
  Size: Integer;
  N: byte;
  Tint: Byte;
begin
  { Deduce the size of the memory for the palette Array work area }
  Size := (256 * SizeOf(TRGBQuad));

  GetMem(QRGBList, Size);                   { Create work area in memory }
  try
    FillChar(QRGBList^, Size, 0);           { initialise to zeros }
    with QRGBList^ do
    begin
      { Generate the palette (i.e. 256 colors) }
{$R-}
      { Generate each of the 128 TPaletteEntries }
      for N := 0 to 127 do
      begin
        Tint  := ((N and 1) shl 5);
        RGBQIndex[N+64].rgbRed := ((N and 96) shl 1) or Tint;
        RGBQIndex[N+64].rgbGreen := ((N and 24) shl 3) or Tint;
        RGBQIndex[N+64].rgbBlue := ((N and 6) shl 5) or Tint;
      end;
    end;
{$R+}
    Result := QRGBList;     { Return Pointer }
  except
    raise EDIBException.Create('Error creating default palette');
  end;
end;

(*
**  Creates a 256 colour DIB frame for a 256 colour DIB
*)
function TDIB.Get256ColDIBFrame(const BWidth, BHeight: integer): Pointer;
var
  FullSize: LongInt;
  BitSize: LongInt;                           { BITS size }
  PalSize: Integer;                           { Actual Palette Size }
  TmpBMPInfo: TBitMapInfo;                    { Bitmap Information Header }
  PPal: PRGBQLIST;
  GMemPointer: Pointer;                       { Pointer to DIB Frame }
begin
  Result := nil;
  PalSize := (255 * SizeOf(TRGBQuad));        { One TRGBQuad in TBitampInfo! }
  BitSize := Align_DWord(LongInt(BWidth)) * LongInt(BHeight);{ Only one byte entry for 8 bit }
  FullSize := SizeOf(TBitmapInfo) +
                     BitSize + PalSize;       { One palette entry as part of TBitMapInfo }

  PPal := MakeQDefaultPaletteArray;           { Make our 128 colour palette }

  try
    with TmpBMPInfo.bmiHeader do
    begin
      biSize := SizeOf(TBitmapInfoHeader);    { Size of Header structure }
      biWidth := BWidth;                      { DIB Width }
      biHeight := BHeight;                    { DIB Height }
      biPlanes := 1;                          { Windows only supports single plane DIBs }
      biBitCount := 8;                        { 256 colours }
      biCompression := BI_RGB;                { No Compression }
      biSizeImage := BitSize;                 { Image Bits Size }
      biXPelsPerMeter := 0;
      biYPelsPerMeter := 0;
      biClrUsed := 256;                       { Colours used }
      biClrImportant := 0;                    { Colours important; 0 = all }
    end;

    FMemHandle := GlobalAlloc(GMEM_MOVEABLE,
      FullSize);                              { Grab Memory for it }
    GMemPointer := GlobalLock(FMemHandle);    { Convert to a pointer }

    try
      { Copy the Bitmap file information to image frame }
      HMemCpy(GMemPointer, @TmpBMPInfo, SizeOf(TmpBMPInfo));

      HMemCpy(OffsetPointer(GMemPointer, SizeOf(TBitmapInfo) - SizeOf(TRGBQuad)),
                       PPal, PalSize);        { Copy Palette away }
    except
      Release256Frame;                        { Release Memory if error }
    end;

  finally
    FreeMem(Pointer(PPal), PalSize);          { Free up palette work area }
  end;
  Result := GMemPointer;                      { Return Pointer to BMP file header }
end;

(*
**  Routine to make a copy of the DIB into a Delphi Device Dependant Bitmap
**   8 Bits per pixel = 256 colours
**  24 Bits per pixel = 16 million colours
**  Does the Delphi trick!  Supplies a BMP that is compatible with the systems
**  colour depth!
*)
procedure TDIB.MakeDelphiBMPFromDIB(TBmp: TBitmap; ForceStatusBar: boolean);
var
  ShowStatusBar: boolean;
begin
  if DIBHeaderPointer <> nil then                   { Only if we've a DIB }
  begin
    { If this DIB is 24 bit and the system only supports 8 bit, then reduce colours }
    if (FDIBColourBits > 8) and (FSysColBits <= 8) then
    begin
      ShowStatusBar := False;
      { Check for Large Large DIB .... }
      if (ForceStatusBar) or (Longint(FDIBWidth) * LongInt(FDIBHeight) > 100000) then
        ShowStatusBar := True;
      { Create Delphi DDB }
      CreateBMPFromDIB(ReduceTo256Cols(ShowStatusBar), TBmp);
      Release256Frame;                              { Release 256 colour DIB space}
    end else
      CreateBMPFromDIB(DIBHeaderPointer, TBmp);       { Create Delphi DDB }
    ReleaseDIBLock;                                 { Allow Windows to move it }
  end;
end;

(*
**  Function to copy our DIB to the Clipboard
*)
function TDIB.CopyDIBToClipboard: boolean;
begin
  Result := CopyToClipboard(DIBInfoPointer);
  ReleaseDIBLock;
end;

(*
**  Function to Copy a DIB to the Clipboard.  DIBPtr points to a TBitmapInfo
**  structure.
*)
function TDIB.CopyToClipboard(DIBPtr: Pointer): boolean;
var
  GMemHandle: THandle;
  GPointer: Pointer;
begin
  Result := False;
  try
    OpenClipboard(Application.Handle);              { Open Channel to Clipboard }
    EmptyClipBoard;                                 { Empty previous contents }
    GMemHandle := GlobalAlloc(GMEM_MOVEABLE, FDIBSize);{ Get Memory }
    try
      GPointer := GlobalLock(GMemHandle);           { Get pointer & lock }

      hMemCpy(GPointer, DIBPtr, FDIBSize);          { Copy data to clipboard memory }
      if SetClipboardData(CF_DIB, GMemHandle) <> 0 then{ Give Clipboard the data }
        Result := True;
    finally
      GlobalUnlock(GMemHandle);                     { Unlock the memory }
      CloseClipBoard;                               { Clipboard now owns memory }
    end;
  finally
  end;
end;

(*
**  Routine to create a Delphi DDB from a Memory DIB.  Assumes that the DIB
**  is colour depth compatible with the system display capabilities
**  THE LOADFROMSTREAM METHOD SOMETIMES DOES NOT RELEASE ALL THE APPROPRIATE
**  MEMORY FROM THE BITMAP UNDER CERTAIN CIRCUMSTANCES!!!!
)
procedure TDIB.CreateBMPFromDIB(DIBPtr: Pointer; var TBMP: TBitMap);
var
  BMPStream: TMemoryStream;
begin
  BMPStream := TMemoryStream.Create;
  try
    WriteDIBToStream(BMPStream, DIBPtr);            { Write Passed DIB to stream }
    BMPStream.Seek(0, 0);                           { Return to start of stream }
    TBMP.FreeImage;
    TBMP.LoadFromStream(BMPStream);                 { Create BMP from stream }
  finally
    BMPStream.Free;
  end;
end;

(*
**  Procedure to Create a Delphi TBitMap bits, from a DIB
**  Routine needs to set the currently focused windows palette to the
**  DIBs palette before creating the DDB.
*)
procedure TDIB.CreateBMPFromDIB(DIBPtr: Pointer; var TBMP: TBitMap);

  { Function to return a pointer to the BITS of a DIB }
  function GetDIBitsPointer(DIBPtr: PBitMapInfoHeader): Pointer;
  var
    RGBSize: integer;
  begin
    RGBSize := DIBPtr^.biClrUsed * SizeOf(TRGBQuad);{ Size of RGB table }

    if RGBSize = 0 then
      RGBSize := (1 shl DIBPtr^.biBitCount) * SizeOf(TRGBQuad);

    GetDIBitsPointer := Pointer(LongInt(DIBPtr) + DIBPtr^.biSize + RGBSize);
  end;

var
  DC: HDC;
  POld: HPalette;
  Focus: HWnd;
  TmpBMPInfo: PBitmapInfo;                      { Pointer to TBitmapInfo }
begin
  Focus := GetFocus;                            { Get Currently Focused Form }
  DC := GetDC(Focus);                           { Make into DC }

  TmpBMPInfo := DIBPtr;                         { Point at DIB header }

  if TmpBMPInfo^.bmiHeader.biBitCount <= 8 then { 256 colours have a palette }
    TBmp.Palette := PaletteFromW3DIB(TmpBMPInfo^);{ Give Target BMP Palette }

  POld := SelectPalette(DC, TBMP.Palette, False);{ Select New Palette into it }
  try
    RealizePalette(DC);                         { Realize it }

    { Create the new bitmap from the DIB }
    TBMP.Handle :=
      CreateDIBitMap(DC, TBitMapInfoHeader(DIBPtr^), cbm_Init,
                     GetDIBitsPointer(DIBPtr), TBitMapInfo(DIBPtr^),
                     dib_RGB_Colors);
  finally
    SelectPalette(DC, POld, False);             { Return original palette }
    ReleaseDC(Focus, DC);                       { Release DC }
  end;
end;

(*
**  Routine to dither a dib scanline.  This is all inline code for speed,
**  ignoring the repetitive nature of some of the blocks of code.
**  There are two lines from the source DIB passed in, each pixel represented
**  by BGR bytes * the width of the DIB.
**  The routine is based on a modified Floyd/Steinberg colour reduction
**  The error between the actual RGB value and the Palette equivalent is
**  spread to the pixel to the right, and the pixel below.
**  The second line is then passed in (externally) as the new line one,
**  thereby ensuring that the errors added on the previos second line
**  are included in the next line.   Errors may be positive OR negative.
*)
procedure TDIB.ConvertRGBLine(RGBBuffer, RGBBuffer2, DestPointer: Pointer;
                              LineWidth: integer); assembler;
var
 Red,
 Blue,
 Green: integer;
 RedE,
 BlueE,
 GreenE: integer;
asm
    xor   ax, ax
    mov   [Red], ax
    mov   [Green], ax
    mov   [Blue], ax                            { Clear Red, Green & Blue storage }

    {
      RGB values are stored BGR in the DIB
    }
    les   di,[RGBBuffer]                        { Load ES:DI to point at Line 1 }
    xor   bx, bx
    mov   bl, es:[di]
    mov   [Blue.Byte[0]], bl

    mov   bl, es:[di+1]
    mov   [Green.Byte[0]], bl

    mov   bl, es:[di+2]
    mov   [Red.Byte[0]], bl

    mov   cx, [LineWidth]                       { Loop for width of line }
    {
      for i:= LineWidth downto 1 do
    }
@@Start:
    push  cx                                    { Stack Pixel Count }

    {
      Here we calculate the 'Tint' value.

      Tint := (Red and 63) + (Green and 63) + (Blue and 63)
      if (Red <> 0) and (Green <> 0) and (Blue <> 0) and (Tint > 93) then
        Tint := 1
      else
        Tint := 0
    }
    mov   al, [Red.Byte[0]]                     { Red Value }
    and   al, 00111111b
    je    @@tint_zero
    mov   ah, al

    mov   al, [Green.Byte[0]]
    and   al, 00111111b
    je    @@tint_zero
    adc   ah, al

    mov   al, [Blue.Byte[0]]
    and   al, 00111111b
    je    @@tint_zero
    adc   ah, al

    cmp   ah, 05eh                              { Tint sum less than 94? }
    jc    @@tint_zero                           { Yes - Tint will be 0 }

    mov   al, 1                                 { No - Tint will be 1 }
    jmp   @@NoTint

@@tint_zero:
    xor   al, al

@@NoTint:
    xor   ah, ah
    push  ax                                    { Stack Tint value }


    {
      Now we calculate the Palette entry for the RGB values retrieved.
      This is done by taking the top two bits from the RGB values and arranging
      them into a single byte, such that Red becomes bits 5 & 6, green
      becomes bits 3 and 4, blue becomes bits 1 and 2 and the Tint is bit zero.
      Add 64 because the colours in our generated palette start at position 64.

      pRed := (Red and 192) div 2
      pGreen := (Green and 192) div 8
      pBlue := (Blue and 192) div 32
      PalEntry := 64 + (pRed or pGreen or pBlue or Tint)
    }

    mov   dl, [Red.Byte[0]]
    and   dl, 11000000b
    shr   dl, 1                                 { Bits 5 and 6 }

    mov   bl, [Green.Byte[0]]
    and   bl, 11000000b
    shr   bl, 3                                 { Bits 3 and 4 }

    mov   cl, [Blue.Byte[0]]
    and   cl, 11000000b
    shr   cl, 5                                 { Bits 1 and 2 }


    or    dl, bl                                { Red or Green }
    or    dl, cl                                { R or G or Blue }
    or    dl, al                                { R or G or B of Tint }
    adc   dl, 040h                              { Add in 63 }

    {
      Here we save the new 8 bit palette entry into the target 256 colour BMP
    }
    les   di, [DestPointer]                     { Point ES:DI at Destination }

    mov   es:[di], dl                           { Save Pal Entry 8 bit pixel }

    {
      DestPointer := OffsetPointer(DestPointer, 1)
    }


    mov   dx, 1                 { Skip One Byte }


    {
      Routine to bump Pointers, by offset passed in in DX, whilst ensuring that
      the segment/selector value skips segment boundaries correctly.
      Destroys BX, AX, CX, DX
    }
    xor   bx, bx
    xchg  ax, di                { Offset of pointer into AX }
    add   ax, dx                { Add Offset }
    adc   bx, 0                 { Add in carry to Selector of pointer }
    xchg  ax, di                { Offset handled, now do selector }
    mov   cx, OFFSET __AHSHIFT
    shl   bx, cl
    push  es
    pop   dx
    add   bx, dx
    push  bx
    pop   es

    mov   [DestPointer.Word[0]], di
    mov   [DestPointer.Word[2]], es             { Save new pointer value }

    {
      RGBBuffer := OffsetPointer(RGBBuffer, 3)
    }
    les   di,[RGBBuffer]                        { Get pointer to current RGBs }

    mov   dx, 3

    {
      Routine to bump Pointers, by offset passed in in DX, whilst ensuring that
      the segment/selector value skips segment boundaries correctly.
      Destroys BX, AX, CX, DX
    }
    xor   bx, bx
    xchg  ax, di                { Offset of pointer into AX }
    add   ax, dx                { Add Offset }
    adc   bx, 0                 { Add in carry to Selector of pointer }
    xchg  ax, di                { Offset handled, now do selector }
    mov   cx, OFFSET __AHSHIFT
    shl   bx, cl
    push  es
    pop   dx
    add   bx, dx
    push  bx
    pop   es

    mov   [RGBBuffer.Word[0]],di
    mov   [RGBBuffer.Word[2]],es                { Save new pointer }

    {
      Here we get the Tint value back, and make sure it is adjusted to be in
      bit 5.
      This ensures we can reconstruct RGB values from a Palette entry.  These
      reconstructed values will only use bits 5,6 and 7.  Therefore the error
      is the difference between the Real Red and the Palette generated Red.
    }

    pop   cx
    shl   cl,   5

    {
      Here we calculate the error between the actual RGB value retrieved, and
      the RGB value yielded by the translation to the palette entry

      RedE := (Red - integer(((Red and 192) or Tint))) div 2
      GreenE := (Green - integer(((Green and 192) or Tint))) div 2
      BlueE := (Blue - integer(((Blue and 192) or Tint))) div 2
    }
    xor   dh, dh
    mov   dl,   [Red.Byte[0]]
    mov   bx,   dx                              { Store Red value in BX and DX }
    and   dx,   11000000b                       { Only take notice of top 2 bits }
    or    dx,   cx                              { Set the Tint bit as appropriate }
    shr   bx,   1
    shr   dx,   1                               { Divide both Reds by two }
    sub   bx,   dx                              { Subtract the Palette red from }
    mov   [RedE], bx                            { the real Red and Save in RedE }

    xor   dh,   dh
    mov   dl,   [Green.Byte[0]]                 { Same for Green }
    mov   bx,   dx
    and   dx,   11000000b
    or    dx,   cx
    shr   dx,   1
    shr   bx,   1
    sub   bx,   dx
    mov   [GreenE], bx

    xor   dh,   dh
    mov   dl,   [Blue.Byte[0]]                  { Same for Blue }
    mov   bx,   dx
    and   dx,   11000000b
    or    dx,   cx
    shr   dx,   1
    shr   bx,   1
    sub   bx,   dx
    mov   [BlueE], bx

    {
      Here we will add the divided by 2 error to the pixel to the right, unless
      it is the last pixel in the line - there won't be a pixel to the right.

      if i < LineWidth then
    }

    pop   dx
    push  dx                                  { Counter around the loop }
    dec   dx                                  { On last Pixel in line? }
    jz    @@NoRightPixel                      { Yes - can't do pixel to right! }

    {
      Spread the error to the pixel on the right.
      Check that the addition of the error has not made the new value grater
      than 255 or less than zero and save away for next time round.
      RGB values in a DIB are actually held in BGR order!

      GetRGBPixel(RGBPointer, Red, Green, Blue)
      Red := Red + RedE
      if Red < 0 then
        Red := 0
      else if Red > 255 then
        Red := 255
      Green := Green + GreenE
      if Green < 0 then
        Green := 0
      else if Green > 255 then
        Green := 255
      Blue := Blue + BlueE
      if Blue < 0 then
        Blue := 0
      else if Blue > 255 then
        Blue = 255

    }

    xor   dh, dh
    mov   dl, es:[di]                         { DX now contains Blue word }
    add   dx, bx                              { BX contains Blue error from above }

    cmp   dx, 0ffh
    jle   @@40
    mov   dx, 0ffh

@@40:
    cmp   dx, 0
    jge   @@41
    xor   dx, dx

@@41:
    mov   [Blue.Byte[0]], dl                  { Save it back }


    mov   dl, es:[di+1]                       { Same for Green }
    xor   dh, dh
    add   dx, [GreenE]
    cmp   dx, 0ffh
    jle   @@42
    mov   dx, 0ffh

@@42:
    cmp   dx, 0
    jge   @@43
    xor   dx, dx

@@43:


    mov   [Green.Byte[0]], dl

    mov   dl, es:[di+2]                       { Same for Red }
    xor   dh, dh
    add   dx, [RedE]
    add   dx, cx

    cmp   dx, 0ffh
    jle   @@44
    mov   dx, 0ffh

@@44:
    cmp   dx, 0
    jge   @@45
    xor   dx, dx

@@45:


    mov   [Red.Byte[0]], dl


@@NoRightPixel:
    {
      end;                                    { End of 'if' block
    }

    {
      Here we spread the error to the pixel below the current one.

      GetRGBPixel(RGBBuffer2, RedB, GreenB, BlueB)
      RedB := RedB + RedE
      if RedB < 0 then
        RedB := 0
      else if RedB > 255 then
        RedB := 255
      GreenB := GreenB + GreenE
      if GreenB < 0 then
        GreenB := 0
      else if GreenB > 255 then
        GreenB := 255
      BlueB := BlueB + BlueE
      if BlueB < 0 then
        BlueB := 0
      else if BlueB > 255 then
        BlueB := 255
      PutRGBPixel(RGBPointer2, BRed, BGreen, BBlue);

    }

    les   di,[RGBBuffer2]

    xor   dh, dh
    mov   dl, es:[di]                         { DX is Blue word }
    add   dx, [BlueE]                         { Add in error }

    cmp   dx, 0ffh
    jle   @@46
    mov   dx, 0ffh

@@46:
    cmp   dx, 0
    jge   @@47
    xor   dx, dx

@@47:

    mov   es:[di], dl                         { Save result back }

    mov   dl, es:[di+1]                       { Same for Green }
    xor   dh, dh
    add   dx, [GreenE]

    cmp   dx, 0ffh
    jle   @@48
    mov   dx, 0ffh

@@48:
    cmp   dx, 0
    jge   @@49
    xor   dx, dx

@@49:

    mov   es:[di+1], dl


    mov   dl, es:[di+2]                       { Same for Red }
    xor   dh, dh
    add   dx, [RedE]

    cmp   dx, 0ffh
    jle   @@50
    mov   dx, 0ffh

@@50:
    cmp   dx, 0
    jge   @@51
    xor   dx, dx

@@51:

    mov   es:[di+2], dl


    {
      Move the line below buffer pointer to the next set of RGB values
    }

    mov   dx, 3

    {
      Routine to bump Pointers, by offset passed in in DX, whilst ensuring that
      the segment/selector value skips segment boundaries correctly.
      Destroys BX, AX, CX, DX
    }
    xor   bx, bx
    xchg  ax, di                { Offset of pointer into AX }
    add   ax, dx                { Add Offset }
    adc   bx, 0                 { Add in carry to Selector of pointer }
    xchg  ax, di                { Offset handled, now do selector }
    mov   cx, OFFSET __AHSHIFT
    shl   bx, cl
    push  es
    pop   dx
    add   bx, dx
    push  bx
    pop   es

    mov   [RGBBuffer2.Word[0]],di
    mov   [RGBBuffer2.Word[2]],es             { Save away }


    {
      Check for end of a line.
    }
    pop   cx                                  { Our stacked line poition counter }
    dec   cx                                  { Decrement it }
    jnz   @@Start                             { Have we finished? }

end;

(*
**  Routine to reduce a 24 bit image to a 256 colour image
*)
function TDIB.ReduceTo256Cols(StatusBar: boolean): Pointer;
var
  SrcDIB: Pointer;                          { Pointer to Source 24 bit DIB }
  PFileHeader : PBitMapFileHeader;          { Pointer to  Target Structure }
  DestBMP: Pointer;                         { Pointer to TBITMAPINFO structure }
  BitSize: LongInt;                         { Target 256 colour DIB Bits Size }
  RGBLineBuffer: Pointer;                   { 24 bit DIB first scanline buffer }
  RGBLine2Buffer: Pointer;                  { 24 bit DIB second scanline buffer }
  ScanLine: integer;                        { Counter for current scanline }
  BMPRGBWidth: word;                        { Number of bytes for RGB scanline width }
  SrcAligned,                               { DWord aligned 24 bit scanline width }
  Aligned: longint;                         { DWord aligned 8 bit scanline width }
  ModValue: integer;                        { Divisor for status bar steps }
  CountTo: integer;                         { Count for Loop scanlines to process }
begin
  Result := nil;
  if FDIBColourBits <> 24 then
    raise EDIBException.Create('CANNOT reduce less than a 24 bit colour DIB');

  CountTo := FDIBHeight - 2;                { Number of times round loop }

  if StatusBar then                         { Need a status bar? }
  begin
    TStatusDialogue.CreateStatusDialogue(FStatusForm, stProgressBar,
        'Reducing Colours to match system capability');  { Create Dialogue form }
    FStatusForm.StatusGauge.MaxValue := CountTo;
    ModValue := (FDIBHeight div 8);         { Only Want 8 steps }
    CentreForm(nil, TForm(FStatusForm));    { Centre it  }
    FStatusForm.Show;                       { Let's see it then }
    Application.ProcessMessages;
  end;

  { Variables to help speed up loop later on }
  Aligned := Align_DWord(FDIBWidth);
  SrcAligned := Align_DWord(FDIBWidth * SizeOf(TRGBTriple));
  BMPRGBWidth := FDIBWidth * SizeOf(TRGBTriple);

  SrcDIB := DIBBitsPointer; { SrcDIB now points at start of source bitmap }

  PFileHeader := Get256ColDIBFrame(FDIBWidth, FDIBHeight);

  { Point at Start of DIB bits information }
  DestBMP := OffsetPointer(PFileHeader, SizeOf(TBitmapInfo) + 255 * SizeOf(TRGBQuad));

  BitSize := Align_DWord(LongInt(FDIBWidth)) * LongInt(FDIBHeight);{ Only one byte entry for 8 bit }

  { Get the two line buffer areas for the RGB values }
  GetMem(RGBLineBuffer, BMPRGBWidth);
  FillChar(RGBLineBuffer^, BMPRGBWidth, 0);
  GetMem(RGBLine2Buffer, BMPRGBWidth);
  FillChar(RGBLine2Buffer^, BMPRGBWidth, 0);

  try

    hMemCpy(RGBLineBuffer, SrcDIB, BMPRGBWidth);  { Get First Line of DIB }
    SrcDIB := OffsetPointer(SrcDIB, SrcAligned);  { Point to second line of DIB }
    hMemCpy(RGBLine2Buffer, SrcDIB, BMPRGBWidth); { Get Second Line of DIB }

    { Convert first line straight into Destination }
    ConvertRGBLine(RGBLineBuffer, RGBLine2Buffer, DestBmp, FDIBWidth);
    DestBmp := OffsetPointer(DestBmp, Aligned);   { Move pointer to next dest line }

    { Copy Line 2 (which has errors added to it) to be the next line 1 }
    hMemCpy(RGBLineBuffer, RGBLine2Buffer, BMPRGBWidth);

    for ScanLine := 0 to CountTo do               { Loop for rest of scan lines }
    begin
      hMemCpy(RGBLine2Buffer, SrcDIB, BMPRGBWidth);{ Get next line into line 2 buffer }
      SrcDIB := OffsetPointer(SrcDIB, SrcAligned); { Point at next Source line }

      { Convert line to destination }
      ConvertRGBLine(RGBLineBuffer, RGBLine2Buffer, DestBmp, FDIBWidth);

      DestBmp := OffsetPointer(DestBmp, Aligned); { Point to next Destination line }

      { Copy Line 2 (which has errors added to it) to be the next line 1 }
      hMemCpy(RGBLineBuffer, RGBLine2Buffer, BMPRGBWidth);

      if StatusBar then                           { Faster when no status bar }
        if ((ScanLine mod ModValue) = 0) or (ScanLine = CountTo) then
          FStatusForm.StatusGauge.Progress := ScanLine;
    end;

  finally
    FreeMem(RGBLineBuffer, BMPRGBWidth);
    FreeMem(RGBLine2Buffer, BMPRGBWidth);         { Free RGB line buffers memory }
    ReleaseDIBLock;                               { Allow Windows to move memory }
    if Assigned(FStatusForm) then                 { Free Status Dialogue memory }
    begin
      FStatusForm.Free;
      FStatusForm := nil;
    end;
  end;
  Result := PFileHeader;                          { Return Pointer to new DIB }
end;

(*
**  Function to increase a 256 colour DIB to 24 bit.  This is needed because
**  JPEG can only handle 24 bit images.
*)
function TDIB.IncreaseTo24bit(DIBPointer: Pointer; StatusBar: boolean): THandle;
var
  PBmInfo: PBitmapInfoHeader;               { Convert pointer to Info pointer }
begin
  PBmInfo := DIBPointer;                    { Establish pointer to Info Header }
  Result := 0;                              { Assume Error }

  if PBmInfo^.biSize <>
    SizeOf(TBitmapInfoHeader) then          { Pointing at rubbish }
      raise EDIBException.Create('Source pointer not valid');

  case PBmInfo^.biBitCount of
    24:
      Result := GlobalPtrHandle(DIBPointer);{ 16M colour is fine as it is }
    8, 4:                                   { 256 or 16 colour }
      Result := ExpandTo24bits(DIBPointer,
        True);                              { Return Handle to new 24 bit DIB }
    1:                                      { We won't bother with B & W }
      raise EDIBException.Create('Cannot handle black and white DIBS');
    else
      raise EDIBException.Create('Unknown DIB colour depth');
  end;

end;

(*
**  Routine to convert the current DIB to full 24 bit colour depth.
**  Discards old DIB on completion.
*)
procedure TDIB.ConvertTo24bit(StatusBar: boolean);
var
  NewDIB: THandle;                          { Memory Handle for new Dib }
begin
  case DIBColourBits of
    24:
      exit;                                 { Already 24 bit }
    8, 4:
      begin
        NewDIB :=
          IncreaseTo24Bit(DIBHeaderPointer,
            StatusBar);                     { Convert up to 24 bit }
        if NewDIB <> 0 then                 { Worked Okay }
        begin
          ReleaseDIBLock;
          SetDIBHandle(NewDIB);             { Set ourselves up }
        end else
          raise EDIBException.Create('ERROR during conversion');
      end;
    1:
      raise EDIBException.Create('Cannot handle Black & White images');
    else
      raise EDIBException.Create('Unknown colour depth');
  end;
end;

(*
**  Routine to convert the current DIB down to 8 bit colour depth.
**  Discards old DIB on completion.
*)
procedure TDIB.ConvertTo8bit(StatusBar: boolean);
var
  NewDIB: THandle;                          { Memory Handle for new Dib }
begin
  case DIBColourBits of
    24:
      begin
        ReduceTo256Cols(StatusBar);         { Convert down to 8 bit }
        if FMemHandle <> 0 then             { Worked Okay }
        begin
          GlobalUnlock(FMemHandle);         { Unlock 256 colour DIB }
          NewDIB := FMemHandle;             { CLEAR Releases this! }
          FMemHandle := 0;
          SetDIBHandle(NewDIB);             { Set ourselves up }
        end else
          raise EDIBException.Create('ERROR during conversion');
      end;
    8:
      exit;                                 { Already 8 bit }
    4:
      raise EDIBException.Create('Function cannot handle 16 colour images');
    1:
      raise EDIBException.Create('Function cannot handle Black & White images');
    else
      raise EDIBException.Create('Unknown colour depth');
  end;
end;

(*
**  Function to return a GlobalMemory Handle pointing to new 24 bit version
**  of the passed in 8 or 4 bit DIB i.e.  256 colour or 16 colour
*)
function TDIB.ExpandTo24Bits(SrcPtr: Pointer; Progress: boolean): THandle;

  (*
  **  Function to fill in a TBitmapInfoHeader in situ
  *)
  procedure Fill24DIBHeader(DIBPtr: Pointer; Width, Height: integer;
                            ImageSize: Longint);
  var
    DIBInfo: PBitMapInfoHeader;
  begin
    DIBInfo := DIBPtr;
    with DIBInfo^ do
    begin
      biSize := SizeOf(TBitmapInfoHeader);  { Size of Header structure }
      biWidth := Width;                     { DIB Width }
      biHeight := Height;                   { DIB Height }
      biPlanes := 1;                        { Windows only supports 1 plane DIBs }
      biBitCount := 24;                     { 16M colours }
      biCompression := BI_RGB;              { No Compression }
      biSizeImage := ImageSize -
        SizeOf(TBitmapInfoHeader);          { Image Bits Size }
      biXPelsPerMeter := 0;
      biYPelsPerMeter := 0;                 { Always Zero }
      biClrUsed := 0;                       { Colours used }
      biClrImportant := 0;                  { Colours important; 0 = all }
    end;
  end;

var
  PBInfo: PBitmapInfoHeader;                { Convert pointer to Info pointer }
  PalPtr: PPaletteEntry;                    { Pointer to Palette }
  DestPtr: Pointer;                         { Destination Pointer }
  GlobMem: THandle;                         { Memory Handle to global Memory }
  PBits: PByte;                             { Pointer to Source data bits byte }
  PalOffs: Pointer;                         { Pointer To Start of palette }
  SrcScanOffs: integer;                     { Offset to next Source scanline bytes }
  DstScanOffs: integer;                     { Offset to next Dest scanline bytes }
  SrcScanLineWidth: Longint;                { Aligned width of scanline }
  DstScanLineWidth: Longint;                { Aligned width of scanline }
  ImgSize: LongInt;                         { New Image Size }
  i,                                        { width pixel counter }
  j: integer;                               { Scanline counter }
  Padding: array[0..4] of byte;             { Alignment padding bytes }
begin
  FillChar(Padding, SizeOf(Padding), 0);    { Fill padding with #0 }
  PBInfo := SrcPtr;
  PalOffs := OffsetPointer(SrcPtr,
    SizeOf(TBitmapInfoHeader));             { Offset to start of palette }
  with PBInfo^ do
  begin
    SrcScanLineWidth := ((((biWidth * biBitCount)
          + 31) div 32) * 4);               { Determine Source ScanLine width }
    DstScanLineWidth := (((biWidth + 3)
      div 4) * 4);                          { Determine Dest ScanLine width }
    DstScanOffs := biWidth mod 4;           { Determine padding for dest EOL }

    ImgSize := SizeOf(TBitmapInfoHeader) +
      (DstScanLineWidth * 3 * biHeight);    { Calculate Size of new image }

    GlobMem := GlobalAlloc(GMEM_MOVEABLE, ImgSize);  { Allocate Memory }
    if GlobMem = 0 then                     { Out of memory }
      raise EOutofResources.Create('Out of Memory');

    DestPtr := GlobalLock(GlobMem);         { Get Pointer to new memory }

    Fill24DIBHeader(DestPtr, biWidth, biHeight, ImgSize);

    DestPtr := OffsetPointer(DestPtr,
      SizeOf(TBitmapInfoHeader));           { Point at Dest First data position }

    PBits := OffsetPointer(SrcPtr,
      SizeOf(TBitmapInfoHeader) +
      GetPaletteLength(PBInfo^));           { Point at Start of Source Bits }

    if Progress then                        { We want a progress dialogue }
    begin
      TStatusDialogue.CreateStatusDialogue(
        FStatusForm, stProgressBar,
        'Increasing Colour Depth');         { Create Status Bar }
      with FStatusForm do
      begin
        CentreForm(nil, TForm(FStatusForm));{ Centre it  }
        StatusGauge.MaxValue := biHeight;   { Set max for progress scale }
        StatusGauge.Progress := 0;
        Show;
        Application.ProcessMessages;        { Let's see it then }
      end;
    end;

    try
      if biBitCount = 8 then                { 256 Colour processing }
      begin
        SrcScanOffs :=
          SrcScanLineWidth - biWidth;       { Determine padding bytes at EOL }
        for j := 0 to biHeight - 1 do
        begin
          for i := 0 to biWidth - 1 do
          begin
            PalPtr := OffsetPointer(PalOffs,
              PBits^ * SizeOf(TRGBQuad));   { Calc offset into palette for this index }
            hMemCpy(DestPtr, PalPtr,
              SizeOf(TRGBTriple));          { Put RGB values into destination }
            PBits :=
              OffsetPointer(PBits, 1);      { Point at next source bits byte }
            DestPtr := OffsetPointer(DestPtr,
              SizeOf(TRGBTriple));          { Bump destination pointer }
          end;
          if SrcScanOffs > 0 then
            PBits := OffsetPointer(PBits,
              SrcScanOffs);                 { Skip alignment bytes in source }
          if DstScanOffs > 0 then           { Pad destination scanline }
          begin
            hMemCpy(DestPtr, @Padding,
              DstScanOffs);                 { Write alignment bytes }
            DestPtr := OffsetPointer(DestPtr,
              DstScanOffs);                 { Bump destination pointer }
          end;
          if Progress then                  { If we want a progress bar }
            if ((j mod 8) = 0) then
              FStatusForm.
              StatusGauge.Progress := j;    { Show progress }
        end;
      end else
      begin                                 { 16 Colour processing }
        SrcScanOffs :=
          ((SrcScanLineWidth * 2) -
          Longint(biWidth)) div 2;          { Determine padding bytes at EOL }
        for j := 0 to biHeight - 1 do
        begin
          for i := 1 to (biWidth div 2) do
          begin
            PalPtr := OffsetPointer(PalOffs,
              ((PBits^ and $F0) shr 4) *
              SizeOf(TRGBQuad));            { Calc offset into palette for this index }
            hMemCpy(DestPtr, PalPtr,
              SizeOf(TRGBTriple));          { Put RGB values into destination }
            DestPtr := OffsetPointer(DestPtr,
              SizeOf(TRGBTriple));          { Bump destination pointer }
            PalPtr := OffsetPointer(PalOffs,
              (PBits^ and $0F) *
              SizeOf(TRGBQuad));            { Calc offset into palette for this index }
            hMemCpy(DestPtr, PalPtr,
              SizeOf(TRGBTriple));          { Put RGB values into destination }
            DestPtr := OffsetPointer(DestPtr,
              SizeOf(TRGBTriple));          { Bump destination pointer }
            PBits :=
              OffsetPointer(PBits, 1);      { Point at next source bits byte }
          end;
          if biWidth mod 2 > 0 then         { One extra nibble to process? }
          begin
            PalPtr := OffsetPointer(PalOffs,
              ((PBits^ and $F0) shr 4) *
              SizeOf(TRGBQuad));            { Calc offset into palette for this index }
            hMemCpy(DestPtr, PalPtr,
              SizeOf(TRGBTriple));          { Put RGB values into destination }
            DestPtr := OffsetPointer(DestPtr,
              SizeOf(TRGBTriple));          { Bump destination pointer }
            PBits :=
              OffsetPointer(PBits, 1);      { Point at next source bits byte }
          end;
          if SrcScanOffs > 0 then
            PBits := OffsetPointer(PBits,
              SrcScanOffs);                 { Skip alignment bytes in source }
          if DstScanOffs > 0 then           { Pad destination scanline }
          begin
            DestPtr := OffsetPointer(DestPtr,
              DstScanOffs);                 { Bump destination pointer }
            hMemCpy(DestPtr, @Padding,
              DstScanOffs);                 { Write alignment bytes }
          end;
          if Progress then                  { If we want a progress bar }
            if ((j mod 8) = 0) then
              FStatusForm.
              StatusGauge.Progress := j;    { Show progress }
        end;
      end;
      if Progress then                      { If we want a progress bar }
        FStatusForm.
        StatusGauge.Progress := biHeight;   { Show 100 % progress }
    except
      GlobalUnlock(GlobMem);                { Unlock Global Memory }
      GlobalFree(GlobMem);                  { Release memory on error }
      if Progress then
      begin
        FStatusForm.Free;                   { Free up dialogue }
        FStatusForm := nil;
      end
    end;
    GlobalUnlock(GlobMem);                  { Unlock Global Memory }
    Result := GlobMem;                      { Return Handle to memory }
    if Progress then
    begin
      FStatusForm.Free;                     { Free up dialogue }
      FStatusForm := nil;
    end
  end;
end;

(*
**  Function to change the DIB memory handle to a passed one
*)
procedure TDIB.SetDIBHandle(GlobalHandle: THandle);
var
  DIBPtr: PBitmapInfoHeader;
begin
  DIBPtr := GlobalLock(GlobalHandle);       { Get pointer to memory }
  try
    try
      if DIBPtr^.biSize <>
        SizeOf(TBitmapInfoHeader) then      { Check for valid DIB }
          raise EDIBException.Create(
          'Invalid Memory contents passed');
      ReleaseDIB;                           { Clear out internal data }
      ClearInternals;                       { Incase there wasn't a prev DIB }
      SetInternals(DIBPtr, True);           { Set Pointers etc }
      FDIBHandle := GlobalHandle;           { We own memory now }
    finally
      GlobalUnlock(GlobalHandle);           { Release Pointer/Lock }
    end;
  except
    raise EDIBException.Create('Error Assigning Global Memory to TDIB');
  end;
end;

{$ifdef JPEGHandler}
(*
**  Routine to validate the setting of the JPEG Quality percentage
*)
procedure TDIB.SetJPEGQuality(Value: byte);
begin
  if Value <> FJPEGQuality then
  begin
    if (Value < 0) or (Value > 100) then
      raise EDIBException.Create('Invalid JPEG Quality Setting')
    else
      FJPEGQuality := Value;
  end;
end;
{$endif}

(*
**  Converts a memory DIB to a Windows Metafile which will draw into the
**  passed box size.
*)
class procedure TDIB.ConvertDIBToMetaFile(DIB: TDIB; TargetBox: TPoint;
     var MFile: TMetaFile);
var
  MetaDC: HMETAFILE;
  PBmpInfo: PBitmapInfo;
  BitsPointer: Pointer;
  MetaHandle: THandle;
begin
  MetaDC := CreateMetaFile(nil);            { Create MetaFile DC }
  PBmpInfo := DIB.DIBHeaderPointer;         { Get Pointer to DIB Header }
  BitsPointer := DIB.DIBBitsPointer;        { Get Pointer to DIB Bits }

  { Store the scalable DIB command to the metafile.  This will be played
    back automatically whenever called }
  StretchDIBits(MetaDC,
               0, 0,
               TargetBox.X, TargetBox.Y,
               0, 0,
               DIB.DIBWidth, Dib.DIBHeight,
               BitsPointer,
               PBmpInfo^,
               DIB_RGB_COLORS,
               SRCCOPY);

  MetaHandle := CloseMetafile(MetaDC);      { Get the Metafile Handle }

  with MFile do
  begin
    Handle := MetaHandle;                   { Give Our Metafile over }
    Inch := Screen.PixelsPerInch;           { Scaling factor (normally 96) }
    Height := TargetBox.Y;                  { Target playback size }
    Width := TargetBox.X;
  end;
end;

(*
**  Creates a MetaFile image into a TImage component straight from
**  a file DIB.
*)
class procedure TDIB.CreateMetaImageFromFile(FName: TFileName; var TargetImage: TImage);
var
  DIB: TDIB;
  FStream: TFileStream;
  MFile: TMetaFile;
  Box,
  Img: TDims;
begin
  FStream := TFileStream.Create(
    FName, fmOpenRead);                     { Open Stream for file }
  DIB := TDIB.CreateFromStream(FStream);    { Create DIB From stream }
  MFile := TMetaFile.Create;                { Create Empty MetaFile }
  Box.Width := TargetImage.Width;           { Target Box Size }
  Box.Height := TargetImage.Height;
  Img.Width := DIB.DIBWidth;                { Image Size Aspect/Ratio }
  Img.Height := DIB.DIBHeight;
  ScaleTImage(TargetImage, Img ,Box);       { Scale Target Box }
  try
    ConvertDIBToMetaFile(DIB,
      Point(Box.Width, Box.Height), MFile); { Convert our DIB to a MetaFile }
    TargetImage.Picture.
      MetaFile.Assign(MFile);               { Assign to TImage component }
  finally
    DIB.Free;                               { Free DIB }
    FStream.Free;                           { Free File Stream }
    MFile.Free;                             { Free MetaFile }
  end;
end;

end.
